/// @file Model.h
/// @description
/// @author learnopengl.com - Updated by Josh
/// @lastedit 2024-02-22
/// @license Unlicense - Public Domain

#pragma once

#include <cstdlib>
#include <glad/glad.h>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include <string>
#include <vector>
#include <LearnOpenGL/Mesh.h>
#include <LearnOpenGL/Shader.h>

#include <LearnOpenGL/BoneInfo.h>
#include <map>

namespace LearnOpenGL
{

    uint TextureFromFile(const char *path, const std::string& directory, bool gamma = false);


    class Model
    {
    public:
        // model data
        std::vector<LearnOpenGL::Texture> textures_loaded; // Stores all the textures loaded so far, optimization to make sure textures aren't loaded more than once
        std::vector<LearnOpenGL::Mesh> meshes;
        std::string directory;
        bool gammaCorrection;
        // Expects a filepath to a 3D model
        Model() {}
        Model(const std::string& path, bool gamma = false);

        static Model LoadFromDAEFile(std::string& path, bool gamma = false);
        static Model LoadFromOBJFile(std::string& path);
        // Draws the model, and this all its meshes
        void Draw(Shader& shader);

        std::map<std::string, BoneInfo> GetBoneInfoMap() const { return BoneInfoMap;}
        int GetBoneCount() const { return BoneCounter;}

    protected:
    private:
        std::map<std::string, BoneInfo> BoneInfoMap;
        int BoneCounter = 0;
        // loads a model with supported ASSIMP extensions from file and stores the resulting meshes in the meshes vector
        // NOTE: if you catch a bottleneck on model load, make sure ASSIMP is built in "Release" mode!!
        void loadModel(const std::string& path);

        // processes a node in a recursive manner. Processes each individual mesh located at the node and repeats this process on its children nodes (if any).
        void processNode(aiNode *node, const aiScene* scene);

        Mesh processMesh(aiMesh* mesh, const aiScene* scene);

        // checks all material textures of a given type and loads the textures if they're not loaded yet.
        // the required info is returned as a Texture struct.
        std::vector<Texture> loadMaterialTextures(aiMaterial* mat, aiTextureType type, std::string typeName);

        void SetVertexBoneDataToDefault(Vertex& vertex);

        void SetVertexBoneData(Vertex& vertex, int boneID, float weight);

        void ExtractBoneWeightForVertices(std::vector<Vertex>& vertices, aiMesh* mesh, const aiScene* scene);
    };

}