#include <LearnOpenGL/Bone.h>

namespace LearnOpenGL
{


    Bone::Bone(const std::string &name, int ID, const aiNodeAnim *channel)
    : Name(name), ID(ID), LocalTransform(1.f)
    {
        NumPositions = channel->mNumPositionKeys;

        for (int positionIndex = 0; positionIndex < NumPositions; ++positionIndex)
        {
            aiVector3D aiPosition = channel->mPositionKeys[positionIndex].mValue;
            float timestamp = channel->mPositionKeys[positionIndex].mTime;

            KeyPosition data;

            data.Position = Vector3(aiPosition.x, aiPosition.y, aiPosition.z);
            data.Timestamp = timestamp;
            Positions.push_back(data);
        }

        NumRotations = channel->mNumRotationKeys;
        for (int rotationIndex = 0; rotationIndex < NumRotations; ++rotationIndex)
        {
            aiQuaternion aiOrientation = channel->mRotationKeys[rotationIndex].mValue;
            float timestamp = channel->mRotationKeys[rotationIndex].mTime;
            KeyRotation data;
            data.Orientation = Quaternion(aiOrientation.x, aiOrientation.y, aiOrientation.z, aiOrientation.w);
            data.Timestamp = timestamp;
            Rotations.push_back(data);
        }

        NumScalings = channel->mNumScalingKeys;
        for (int keyIndex = 0; keyIndex < NumScalings; ++keyIndex)
        {
            aiVector3D scale = channel->mScalingKeys[keyIndex].mValue;
            float timestamp = channel->mScalingKeys[keyIndex].mTime;
            KeyScale data;
            data.Scale = Vector3(scale.x, scale.y, scale.z);
            data.Timestamp = timestamp;
            Scales.push_back(data);
        }
    }

    void Bone::Update(float animationTime) {
        Matrix4x4 Translation = InterpolatePosition(animationTime);
        Matrix4x4 Rotation = InterpolateRotation(animationTime);
        Matrix4x4 Scale = InterpolateRotation(animationTime);

        LocalTransform = Translation * Rotation * Scale;
        //LocalTransform = Matrix4x4::FromTRS(Translation, Rotation, Scale)
    }

    Matrix4x4 Bone::GetLocalTransform() const { return LocalTransform;}

    std::string Bone::GetBoneName() const { return Name;}

    int Bone::GetBoneID() const { return ID;}

    int Bone::GetPositionIndex(float animationTime) {
        for (int index = 0; index < NumPositions - 1; ++index)
        {
            if (animationTime < Positions[index+1].Timestamp)
                return index;
        }
        assert(0);
    }

    int Bone::GetRotationIndex(float animationTime) {
        for (int index = 0; index < NumRotations - 1; ++index)
        {
            if (animationTime < Rotations[index + 1].Timestamp)
                return index;
        }
        assert(0);
    }

    int Bone::GetScaleIndex(float animationTime) {
        for (int index = 0; index < NumScalings - 1; ++index)
        {
            if (animationTime < Scales[index+1].Timestamp)
                return index;
        }
        assert(0);
    }

    float Bone::GetScaleFactor(float lastTimestamp, float nextTimestamp, float animationTime) {
        float scaleFactor = 0.f;
        float midwayLength = animationTime - lastTimestamp;
        float framesDiff = nextTimestamp - lastTimestamp;
        scaleFactor = midwayLength / framesDiff;
        return scaleFactor;
    }

    Matrix4x4 Bone::InterpolatePosition(float animationTime) {
        if (1 == NumPositions)
            return Matrix4x4(1.f).Translate(Positions[0].Position);

        int p0Index = GetPositionIndex(animationTime);
        int p1Index = p0Index + 1;
        float scaleFactor = GetScaleFactor(Positions[p0Index].Timestamp,
                                           Positions[p1Index].Timestamp, animationTime);
        Vector3 finalPosition = Vector3::Lerp(Positions[p0Index].Position, Positions[p1Index].Position, scaleFactor);
        return Matrix4x4(1.f).Translate(finalPosition);
    }

    Matrix4x4 Bone::InterpolateRotation(float animationTime) {
        if (1 == NumRotations)
        {
            auto rotation = Rotations[0].Orientation.Normalize();
            return Matrix4x4(rotation);
        }

        int p0Index = GetRotationIndex(animationTime);
        int p1Index = p0Index + 1;
        float scaleFactor = GetScaleFactor(Rotations[p0Index].Timestamp,
                                           Rotations[p1Index].Timestamp, animationTime);
        Quaternion finalRotation = Rotations[p0Index].Orientation.Lerp(Rotations[p1Index].Orientation, scaleFactor);
        finalRotation = finalRotation.Normalize();
        return Matrix4x4(finalRotation);
    }

    Matrix4x4 Bone::InterpolateScaling(float animationTime) {
        if (1 == NumScalings)
        {
            return Matrix4x4(1.f).Scale(Scales[0].Scale);
        }

        int p0Index = GetScaleIndex(animationTime);
        int p1Index = p0Index + 1;

        float scaleFactor = GetScaleFactor(Scales[p0Index].Timestamp,
                                           Scales[p1Index].Timestamp, animationTime);
        Vector3 finalScale = Scales[p0Index].Scale.Lerp(Scales[p1Index].Scale, scaleFactor);
        return Matrix4x4(1.f).Scale(finalScale);
    }
}