#include <LearnOpenGL/VertexArray.h>
#include <fstream>
#include <iostream>
#include <sstream>
#include "J3ML/J3ML.h"

namespace LearnOpenGL
{
    bool useVBO = false;

    void VertexArray::load (const std::string& filename) {
        std::ifstream file(filename);
        if (!file.is_open()) {
            std::cerr << "Couldn't Load Model: " << filename << std::endl;
            //engine->setError(ENGINE_ERROR_CODE::MODEL_NOT_FOUND, true);
        }
        std::vector<Vector3> positions;
        std::vector<Vector2> uvs;
        std::string line;
        while (std::getline(file, line)) {
            std::istringstream stream(line);
            std::string prefix;
            stream >> prefix;

            if (prefix == "v") {
                float x, y, z;
                stream >> x >> y >> z;
                positions.push_back(Vector3(x, y, z));
            } else if (prefix == "vt") {
                float u, v;
                stream >> u >> v;
                uvs.push_back(Vector2(u, v));
            } else if (prefix == "f") {
                unsigned int vertexIndex[3], texCoordIndex[3];
                char slash;
                for (int i = 0; i < 3; ++i) {
                    stream >> vertexIndex[i] >> slash >> texCoordIndex[i];
                    vertexIndex[i]--;
                    texCoordIndex[i]--;
                }

                for (int i = 0; i < 3; ++i) {
                    Vertex vertex;
                    vertex.Position = positions[vertexIndex[i]];
                    vertex.TexCoords = uvs[texCoordIndex[i]];

                    vertices.push_back(vertex);
                    indices.push_back(static_cast<unsigned int>(indices.size()));
                }
            }
        }
        file.close();

        //Copy verts to vram.
        if (useVBO) {
            if (vbo == NULL) {
                GLfloat verts[vertices.size() * 3];
                size_t index = 0;
                for (const Vertex &vertex: vertices) {
                    verts[index++] = vertex.Position.x;
                    verts[index++] = vertex.Position.y;
                    verts[index++] = vertex.Position.z;
                }

                glGenBuffers(1, &vbo);
                glBindBuffer(GL_ARRAY_BUFFER, vbo);
                glBufferData(GL_ARRAY_BUFFER, sizeof(verts), &verts, GL_STATIC_DRAW); //GL_STATIC_DRAW means the VBO won't be changed atall.
                glBindBuffer(GL_ARRAY_BUFFER, 0);
            }
        }

        // TODO: Re-implement BoundingVolume calculation
        //boundingVolume = Collision::calculateBoundingBox(this, {0,0,0}).vertices;

    }

//TODO: Render queue

    void VertexArray::draw() {
        if (!indices.empty()) {
            glEnable(GL_TEXTURE_2D);
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

            glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), vertices[0].TexCoords.ptr());
            if (!useVBO)
                glVertexPointer(3, GL_FLOAT, sizeof(Vertex), vertices[0].Position.ptr());

            if (useVBO) {
                glBindBuffer(GL_ARRAY_BUFFER, vbo);
                glVertexPointer(3, GL_FLOAT, 0, 0);
            }
            glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &indices[0]);

            if (useVBO)
                glBindBuffer(GL_ARRAY_BUFFER, 0);

            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            glDisableClientState(GL_VERTEX_ARRAY);
            glDisable(GL_TEXTURE_2D);
        }

        //This only works for *really* simple shapes.
        //Primarily for the bounding box.
        if (indices.empty()) {
            glBegin(GL_LINES);
            for (int i = 0; i < 4; ++i) {

                auto vertA = vertices[i];
                auto vertB = vertices[(i + 1) % 4];
                auto vertC = vertices[((i+1)%4) + 4];
                auto vertD = vertices[i+4];
                glVertex3f(vertA.Position.x, vertA.Position.y, vertA.Position.z);
                glVertex3f(vertB.Position.x, vertB.Position.y, vertB.Position.z);

                glVertex3f(vertD.Position.x, vertD.Position.y, vertD.Position.z);
                glVertex3f(vertC.Position.x, vertC.Position.y, vertC.Position.z);

                glVertex3f(vertA.Position.x, vertA.Position.y, vertA.Position.z);
                glVertex3f(vertD.Position.x, vertD.Position.y, vertD.Position.z);
            }
            glEnd();
        }
    }

    void VertexArray::drawWireframe() {
        glBegin(GL_LINES);
        for (size_t i = 0; i < indices.size(); i += 2) {
            if (i + 1 < indices.size()) {
                int index1 = static_cast<int>(indices[i]);
                int index2 = static_cast<int>(indices[i + 1]);

                if (index1 >= 0 && index1 < vertices.size() && index2 >= 0 && index2 < vertices.size()) {
                    glVertex3f(vertices[index1].Position.x, vertices[index1].Position.y, vertices[index1].Position.z);
                    glVertex3f(vertices[index2].Position.x, vertices[index2].Position.y, vertices[index2].Position.z);
                }
            }
        }
        glEnd();
    }

    void VertexArray::rotate(const Vector3& angle) {
        float angleX = J3ML::Math::Radians(angle.x);
        float angleY = J3ML::Math::Radians(angle.y);
        float angleZ = J3ML::Math::Radians(angle.z);

        for (Vertex &vertex: vertices) {
            float newX = vertex.Position.x * std::cos(angleZ) - vertex.Position.y * std::sin(angleZ);
            float newY = vertex.Position.x * std::sin(angleZ) + vertex.Position.y * std::cos(angleZ);
            vertex.Position.x = newX;
            vertex.Position.y = newY;
        }

        for (Vertex &vertex: vertices) {
            float newX = vertex.Position.x * std::cos(angleY) + vertex.Position.z * std::sin(angleY);
            float newZ = -vertex.Position.x * std::sin(angleY) + vertex.Position.z * std::cos(angleY);
            vertex.Position.x = newX;
            vertex.Position.z = newZ;
        }

        for (Vertex &vertex: vertices) {
            float newY = vertex.Position.y * std::cos(angleX) - vertex.Position.z * std::sin(angleX);
            float newZ = vertex.Position.y * std::sin(angleX) + vertex.Position.z * std::cos(angleX);
            vertex.Position.y = newY;
            vertex.Position.z = newZ;
        }
    }

    VertexArray VertexArray::getBoundingVolume(Vector3 angle) {
        VertexArray vArray;
        vArray.vertices = boundingVolume;
        vArray.rotate(angle);
        return vArray;

    }

    void VertexArray::erase(const VertexArray& vArray) {
        if (useVBO)
            glDeleteBuffers(1, &vArray.vbo);

        //for (int i = 0; i < engine->world->geometryList.size(); i++) {
            //if (engine->world->geometryList[i].name == vArray.name)
               //engine->world->geometryList.erase(engine->world->geometryList.begin() + i);
    }
}
